{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Time series\n\nInterpolation of a time series\n\nThis example shows how to interpolate a time series using the library.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>This example is not executed because it needs to access data in the\n    e-cloud. But you can run it on binder.</p></div>\n\nIn this example, we consider the time series of MSLA maps distributed by\nAVISO/CMEMS. We start by retrieving the data:\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import datetime\n\nimport cartopy.crs\nimport cartopy.feature\nimport intake\nimport matplotlib.pyplot\nimport numpy\nimport pandas\n\nimport pyinterp.backends.xarray\nimport pyinterp.tests\n\ncat = intake.open_catalog('https://raw.githubusercontent.com/pangeo-data'\n                          '/pangeo-datastore/master/intake-catalogs/'\n                          'ocean.yaml')\nds = cat['sea_surface_height'].to_dask()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "To manage the time series retrieved, we create the following object:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class TimeSeries:\n    \"\"\"Manage a time series composed of a grid stack.\"\"\"\n\n    def __init__(self, ds):\n        self.ds = ds\n        self.series, self.dt = self._load_ts()\n\n    @staticmethod\n    def _is_sorted(array):\n        indices = numpy.argsort(array)\n        return numpy.all(indices == numpy.arange(len(indices)))\n\n    def _load_ts(self):\n        \"\"\"Loading the time series into memory.\"\"\"\n        time = self.ds.time\n        assert self._is_sorted(time)\n\n        series = pandas.Series(time)\n        frequency = set(\n            numpy.diff(series.values.astype('datetime64[s]')).astype('int64'))\n        if len(frequency) != 1:\n            raise RuntimeError(\n                'Time series does not have a constant step between two '\n                f'grids: {frequency} seconds')\n        return series, datetime.timedelta(seconds=float(frequency.pop()))\n\n    def load_dataset(self, varname, start, end):\n        \"\"\"Loading the time series into memory for the defined period.\n\n        Args:\n            varname (str): Name of the variable to be loaded into memory.\n            start (datetime.datetime): Date of the first map to be loaded.\n            end (datetime.datetime): Date of the last map to be loaded.\n\n        Returns:\n            pyinterp.backends.xarray.Grid3D: The interpolator handling the\n            interpolation of the grid series.\n        \"\"\"\n        if start < self.series.min() or end > self.series.max():\n            raise IndexError(\n                f'period [{start}, {end}] out of range [{self.series.min()}, '\n                f'{self.series.max()}]')\n        first = start - self.dt\n        last = end + self.dt\n\n        selected = self.series[(self.series >= first) & (self.series < last)]\n        print(f'fetch data from {selected.min()} to {selected.max()}')\n\n        data_array = ds[varname].isel(time=selected.index)\n        return pyinterp.backends.xarray.Grid3D(data_array)\n\n\ntime_series = TimeSeries(ds)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "The test data set containing a set of positions of different floats is then\nloaded.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def cnes_jd_to_datetime(seconds):\n    \"\"\"Convert a date expressed in seconds since 1950 into a calendar date.\"\"\"\n    return datetime.datetime.utcfromtimestamp(\n        ((seconds / 86400.0) - 7305.0) * 86400.0)\n\n\ndef load_positions():\n    \"\"\"Loading and formatting the dataset.\"\"\"\n    df = pandas.read_csv(\n        pyinterp.tests.positions_path(),\n        header=None,\n        sep=r';',\n        usecols=[0, 1, 2, 3],\n        names=['id', 'time', 'lon', 'lat'],\n        dtype={\n            'id': numpy.uint32,\n            'time': numpy.float64,\n            'lon': numpy.float64,\n            'lat': numpy.float64,\n        },\n    )\n    df.mask(df == 1.8446744073709552e+19, numpy.nan, inplace=True)\n    df['time'] = df['time'].apply(cnes_jd_to_datetime)\n    df.set_index('time', inplace=True)\n    df['sla'] = numpy.nan\n    return df.sort_index()\n\n\ndf = load_positions()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Two last functions are then implemented. The first function will divide the\ntime series to be processed into weeks.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def periods(df, time_series, frequency='W'):\n    \"\"\"Return the list of periods covering the time series loaded in memory.\"\"\"\n    period_start = df.groupby(\n        df.index.to_period(frequency))['sla'].count().index\n\n    for start, end in zip(period_start, period_start[1:], strict=False):\n        start_timestamp = start.to_timestamp()\n        start_timestamp = max(start_timestamp, time_series.series[0])\n        end_timestamp = end.to_timestamp()\n        yield start_timestamp, end_timestamp\n    yield end_timestamp, df.index[-1] + time_series.dt"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "The second one will interpolate the DataFrame loaded in memory.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def interpolate(df, time_series, start, end):\n    \"\"\"Interpolate the time series over the defined period.\"\"\"\n    interpolator = time_series.load_dataset('sla', start, end)\n    mask = (df.index >= start) & (df.index < end)\n    selected = df.loc[mask, ['lon', 'lat']]\n    df.loc[mask, ['sla']] = interpolator.trivariate(\n        {\n            'longitude': selected['lon'].values,\n            'latitude': selected['lat'].values,\n            'time': selected.index.values\n        },\n        interpolator='inverse_distance_weighting',\n        num_threads=0)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Finally, the SLA is interpolated on all loaded floats.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "for start, end in periods(df, time_series, frequency='M'):\n    interpolate(df, time_series, start, end)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Visualization of the SLA for a float.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "float_id = 62423050\nselected_float = df[df.id == float_id]\nfirst = selected_float.index.min()\nlast = selected_float.index.max()\nsize = (selected_float.index - first) / (last - first)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "fig = matplotlib.pyplot.figure(figsize=(10, 5))\nax = fig.add_subplot(111,\n                     projection=cartopy.crs.PlateCarree(central_longitude=180))\nsc = ax.scatter(selected_float.lon,\n                selected_float.lat,\n                s=size * 100,\n                c=selected_float.sla,\n                transform=cartopy.crs.PlateCarree(),\n                cmap='jet')\nax.coastlines()\nax.set_title('Time series of SLA '\n             '(larger points are closer to the last date)')\nax.add_feature(cartopy.feature.LAND)\nax.add_feature(cartopy.feature.COASTLINE)\nax.set_extent([80, 100, 13.5, 25], crs=cartopy.crs.PlateCarree())\nfig.colorbar(sc)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "The image below illustrates the result of the code above:\n\n.. figure:: ../pictures/time_series.png\n    :align: center\n\n    Time series of SLA observed by float #62423050\n    (larger points are closer to the last date)\n\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.12.7"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}